{------------------------------------------------------------------------------------------------

   Unit:      AEGCheckedCombo
   Author:    Scott D. Slater
   Email:     sevenout@jps.net
   Written:   8-1-2000
   Copyright: (c) 2000-2001 Scott D. Slater
  
   DISCLAIMER
      This software is provided 'as-is', without any express or implied warranty. In no
      event will the author be held liable for any damages arising from the use of this
      software.
  
   DESCRIPTION
      This unit provides a data aware combo box where each item in the drop down
      list has a check box (ala TCheckListBox).  This component is based on the component
      called TCheckedComboBox by Tsourinakis Antonis (tsoyran@otenet.gr).
      This unit has been tested extensively under Delphi 4, but should work under Delphi 3-5.
  
   PROPERTIES
      Checked[Index: Integer]: Boolean read GetChecked write SetChecked;
         For each item on the drop down list, Checked is True if a check mark appears
         in the items check box.

      CheckedCount: Integer read FCheckedCount;
         A count of the number of items on the drop down list that are checked.

      SelectState: TCHBSelectState read FSelectState;
         Indicates the current state of the combo box with regards to how many items are
         selected.  ssNone     = No items selected
                    ssOne      = Exactly 1 item is selected
                    ssMultiple = More than 1 but less than All items are selected
                    ssAll      = All items are selected

      Columns: Integer read FColumns write SetColumns default 0;
         Indicates the number of columns of items that should appear on the drop down list.

      DropDownLines: Integer read FDropDownLines write SetDropDownLines default 7;
         Indicates the maximum number of lines that should be displayed before
         scroll bars are shown.

      Loading: Boolean read FLoading default False;
         If True, prevents the LocateIndex and LoadData methods from performing any processing.
         Eg. This can be used as a way to prevent the control from loading data if the LoadData
         method is called from an OnDataChange event.

      ForceAll: Boolean read FForceAll write SetForceAll default False;
         Has no real effect on whether the user can select less that All the items.  Instead
         it forces the display the show "All" and sets the SelectState = ssAll.

      DataSource: TDataSource read GetDataSource write SetDataSource;
         The datasource where the data for the drop down list is coming from.

      DataField: string read GetDataField write SetDataField;
         The data field to display as the drop down items.

      Items: TStrings read FItems write FItems;
         This is the array of items that the user can select on the drop down list.  If the
         DataSource and DataField properties are set and the LoadData method is called, this
         array is automatically filled with data from the dataset.  If you use it as a non-
         data aware component, you can manually or programatically set the items that can be
         selected.

      Sorted: Boolean read FSorted write SetSorted default False;
         TRUE = Items on the drop down list are sorted
         FLASE = Items are not sorted and are left in the order they are added.

      AllowNone: Boolean read FAllowNone write SetAllowNone default False;
         TRUE = Allow the user to select no items on the list.
         FALSE = At least one item must always be selected as long as there are items on the list.
  
   METHODS
      procedure ClearItems;
         Removes all items from the drop down list.  SelectState will be set to ssNone and the
         display will show "None".

      procedure CheckAll;
         Checks all items on the list.  SelectState will be set to ssAll and the display will
         show "All"

      procedure UnCheckAll;
         Unchecks all items on the list.  SelectState will be set to ssNone and the display will
         show "None".  If AllowNone is False, the first item on the list will automatically be
         checked.

      procedure LoadData(Checked: Boolean);
         Loads items on the list from the datasource specified.  The items loaded are derived
         from the values in the DataField property.

      procedure LocateIndex(Index: Integer);
         Cause the DataSource record pointer to be moved to the item specified by Index.

      function IsChecked(Index: Integer): Boolean;
         Returns True if the item specified by Index is checked.  Otherwise it returns False.

      function GetText(Quote: TCHBQuoteType): String;
         Returns a comma delimited string of the selected items.
         The text of the items are quoted based on the specified QuoteType.  Example:
            qtNone   -> Item1,Item2,Item3,Item4
            qtSingle -> 'Item1','Item2','Item3','Item4'
            qtDouble -> "Item1","Item2","Item3","Item4"

      function GetFieldText(Field: String; Quote: TCHBQuoteType): String;
         Returns a comma delimited string of the selected items.  The item text returned is the
         value of the Field within the datasource.  This way you can get a different list of
         values than what is displayed.
         The text of the items are quoted based on the specified QuoteType.  Example:
            qtNone   -> Item1,Item2,Item3,Item4
            qtSingle -> 'Item1','Item2','Item3','Item4'
            qtDouble -> "Item1","Item2","Item3","Item4"

      function GetObjectText(Quote: TCHBQuoteType): String;
         Returns a comma delimited string of the selected items.  The item text returned is the
         value of the Object property of the item list instead of the item text.
         The text of the items are quoted based on the specified QuoteType.  Example:
            qtNone   -> Item1,Item2,Item3,Item4
            qtSingle -> 'Item1','Item2','Item3','Item4'
            qtDouble -> "Item1","Item2","Item3","Item4"

------------------------------------------------------------------------------------------------}

unit AEGCheckedCombo;

{$R-}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, CheckLst, ExtCtrls, Buttons, Db, DbTables, DbCtrls;

type
  TCHBSelectState = (ssNone, ssOne, ssMultiple, ssAll);
  TCHBQuoteType = (qtNone, qtSingle, qtDouble);

  TAEGPopupGrid = class;

  TAEGCheckedCombo = class(TCustomEdit)
  private
    { Private declarations }
    FCanvas: TControlCanvas;
    FFieldLink: TFieldDataLink;
    FTextMargin: Integer;
    FButton: TSpeedButton;
    FAllBtn: TSpeedButton;
    FNoneBtn: TSpeedButton;
    FBtnControl: TWinControl;
    FItems: TStrings;
    FListBox: TAEGPopupGrid;
    FSorted: Boolean;
    FSelectState: TCHBSelectState;
    FCheckedCount: Integer;
    FColumns: Integer;
    FDropDownLines: Integer;
    FActualLines: Integer;
    FLoading: Boolean;
    FDataItems: TStrings;
    FAllowNone: Boolean;
    FForceAll: Boolean;
    function GetDataSource: TDataSource;
    function GetDataField: string;
    procedure SetDataSource(Value: TDataSource);
    procedure SetDataField(const Value: string);
    procedure SetSorted(Value: Boolean);
    procedure SetAllowNone(Value: Boolean);
    procedure SetForceAll(Value: Boolean);
    procedure SetColumns(Value: Integer);
    procedure SetChecked(Index: Integer; Checked: Boolean);
    function GetChecked(Index: Integer): Boolean;
    procedure SetDropDownLines(Value: Integer);
    procedure ItemsChange(Sender: TObject);
    procedure SetEditRect;
    function GetMinHeight: Integer;
    procedure ToggleOnOff(Sender: TObject);
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure NonEditMouseDown(var Message: TWMLButtonDown);
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CANCELMODE;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure DropDown;
    procedure CloseUp;
    procedure UpdateText;
    procedure DoChange;
    procedure SetUnCheckedAll(Sender: TObject);
    procedure SetCheckedAll(Sender: TObject);
  protected
    { Protected declarations }
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function IsChecked(Index: Integer): Boolean;
    property Checked[Index: Integer]: Boolean read GetChecked write SetChecked;
    property CheckedCount: Integer read FCheckedCount;
    property SelectState: TCHBSelectState read FSelectState;
    procedure ClearItems;
    procedure CheckAll;
    procedure UnCheckAll;
    function GetText(Quote: TCHBQuoteType): String;
    function GetFieldText(Field: String; Quote: TCHBQuoteType): String;
    function GetObjectText(Quote: TCHBQuoteType): String;
    procedure LoadData(Checked: Boolean);
    procedure LocateIndex(Index: Integer);
    property Columns: Integer read FColumns write SetColumns default 0;
    property DropDownLines: Integer read FDropDownLines write SetDropDownLines default 7;
    property Loading: Boolean read FLoading default False;
    property ForceAll: Boolean read FForceAll write SetForceAll default False;
  published
    { Published declarations }
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DataField: string read GetDataField write SetDataField;
    property Items: TStrings read FItems write FItems;
    property Sorted: Boolean read FSorted write SetSorted default False;
    property AllowNone: Boolean read FAllowNone write SetAllowNone default False;
    // Inherited properties
    property Anchors;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ImeMode;
    property ImeName;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    // Inherited events
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

{ TAEGPopupGrid }

  TAEGPopupGrid = class(TCheckListBox)
  protected
    FLastIndex: Integer;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  end;

{ TAEGComboButton }

  TAEGComboButton = class(TSpeedButton)
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  end;

procedure Register;

implementation
{$R *.RES}

const
	MINDROPLINES	= 1;
	MAXDROPLINES	= 7;

{TAEGCheckedCombo}

constructor TAEGCheckedCombo.Create (AOwner: TComponent);
begin
   inherited Create(AOwner);
   AutoSize        := False;
   Height          := 21;
   ReadOnly        := True;
   Text            := 'None';
   FDropDownLines  := MINDROPLINES;
   FColumns        := 0;
   FCheckedCount   := 0;
   FSelectState    := ssNone;
   FLoading        := False;
   FSorted         := False;
   FAllowNone      := False;
   FForceAll       := False;

   FFieldLink := TFieldDataLink.Create;
   FFieldLink.Control := Self;

   FDataItems     := TStringList.Create;
   FItems         := TStringList.Create;
   TStringList(FItems).OnChange := ItemsChange;

   FBtnControl          := TWinControl.Create(Self);
   FBtnControl.Width    := 51;
   FBtnControl.Height   := 17;
   FBtnControl.Visible  := True;
   FBtnControl.Parent   := Self;
   FButton := TAEGComboButton.Create(Self);
   FButton.SetBounds(0, 0, 17, FBtnControl.Height);
   FButton.Glyph.Handle := LoadBitmap(0, PChar(32738));
   FButton.Visible      := True;
   FButton.Parent       := FBtnControl;

   FAllBtn := TSpeedButton.Create(Self);
   FAllBtn.SetBounds(17, 0, 17, FBtnControl.Height);
   FAllBtn.Glyph.LoadFromResourceName(HInstance, 'AEGCHKALL');
   FAllBtn.NumGlyphs    := 2;
   FAllBtn.Visible      := True;
   FAllBtn.Parent       := FBtnControl;
   FAllBtn.Hint         := 'Select All';
   FAllBtn.ShowHint     := True;
   FAllBtn.OnClick      := SetCheckedAll;

   FNoneBtn := TSpeedButton.Create(Self);
   FNoneBtn.SetBounds(34, 0, 17, FBtnControl.Height);
   FNoneBtn.Glyph.LoadFromResourceName(HInstance, 'AEGCHKNONE');
   FNoneBtn.NumGlyphs   := 2;
   FNoneBtn.Visible     := FAllowNone;
   FNoneBtn.Parent      := FBtnControl;
   FNoneBtn.Hint        := 'Select None';
   FNoneBtn.ShowHint    := True;
   FNoneBtn.OnClick     := SetUnCheckedAll;
   if not FAllowNone then FBtnControl.Width := 34;

   FListBox                := TAEGPopupGrid.Create(Self);
   FListBox.Parent         := Self;
   FListBox.Visible        := False;
   FListBox.Ctl3D          := False;
   FListBox.Columns        := FColumns;
   FListBox.OnClickCheck   := ToggleOnOff;
end;

destructor TAEGCheckedCombo.Destroy;
begin
   FFieldLink.Free;
   FFieldLink := nil;
   FListBox.Free;
   FItems.Free;
   FDataItems.Free;
   inherited Destroy;
end;

procedure TAEGCheckedCombo.CreateParams(var Params: TCreateParams);
begin
   inherited CreateParams(Params);
   Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN;
end;

procedure TAEGCheckedCombo.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited Notification(AComponent, Operation);
   if (Operation = opRemove) and (FFieldLink <> nil) and (AComponent = DataSource) then
      DataSource := nil;
end;

function TAEGCheckedCombo.GetDataSource: TDataSource;
begin
  Result := FFieldLink.DataSource;
end;

procedure TAEGCheckedCombo.SetDataSource(Value: TDataSource);
begin
  if (Value <> nil) and (Value <> DataSource) then
     FFieldLink.DataSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

function TAEGCheckedCombo.GetDataField: string;
begin
  Result := FFieldLink.FieldName;
end;

procedure TAEGCheckedCombo.SetDataField(const Value: string);
begin
  FFieldLink.FieldName := Value;
end;

procedure TAEGCheckedCombo.Loaded;
begin
   inherited Loaded;
end;

procedure TAEGCheckedCombo.CreateWnd;
begin
	inherited CreateWnd;
   SetEditRect;
end;

procedure TAEGCheckedCombo.SetEditRect;
var
   Loc: TRect;
begin
   Loc.Bottom := ClientHeight + 1;  {+1 is workaround for windows paint bug}
   Loc.Right := FBtnControl.Left - 2;
   Loc.Top := 0;
   Loc.Left := 0;
   SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
end;

function TAEGCheckedCombo.GetMinHeight: Integer;
var
   DC: HDC;
   SaveFont: HFont;
   I: Integer;
   SysMetrics, Metrics: TTextMetric;
begin
   DC := GetDC(0);
   GetTextMetrics(DC, SysMetrics);
   SaveFont := SelectObject(DC, Font.Handle);
   GetTextMetrics(DC, Metrics);
   SelectObject(DC, SaveFont);
   ReleaseDC(0, DC);
   I := SysMetrics.tmHeight;
   if I > Metrics.tmHeight then I := Metrics.tmHeight;
   FTextMargin := I div 4;
   Result := Metrics.tmHeight + FTextMargin + GetSystemMetrics(SM_CYBORDER) * 4 + 1;
end;

procedure TAEGCheckedCombo.WMPaint(var Message: TWMPaint);
var
  PS: TPaintStruct;
  ARect: TRect;
  TextLeft, TextTop: Integer;
  Focused: Boolean;
  DC: HDC;
const
  Formats: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT,
    DT_CENTER or DT_WORDBREAK or DT_EXPANDTABS or DT_NOPREFIX);
begin
  if FCanvas = nil then
  begin
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;

  DC := Message.DC;
  if DC = 0 then DC := BeginPaint(Handle, PS);
  FCanvas.Handle := DC;
  try
    Focused := (GetFocus = Handle) and (not FListBox.Visible);
    FCanvas.Font := Font;
    with FCanvas do
    begin
      ARect := ClientRect;
      Brush.Color := clWindowFrame;
      Brush.Style := bsSolid;
      Brush.Color := Color;
      FillRect (ARect);
      TextTop := FTextMargin - 1;
      if TextTop < 0 then TextTop := 0;
      ARect.Left := ARect.Left + 1;
      ARect.Right := FBtnControl.Left - 1;
      TextLeft := FTextMargin;
      if Focused then
      begin
        Brush.Color := clHighlight;
        Font.Color := clHighlightText;
        ARect.Top := ARect.Top + 1;
        ARect.Bottom := ARect.Bottom - 1;
      end;
      ExtTextOut(FCanvas.Handle, TextLeft, TextTop, ETO_OPAQUE or ETO_CLIPPED, @ARect,
        PChar(Text), Length(Text), nil);
      if Focused then
        DrawFocusRect(ARect);
    end;
  finally
    FCanvas.Handle := 0;
    if Message.DC = 0 then EndPaint(Handle, PS);
  end;
end;

procedure TAEGCheckedCombo.WMSize(var Message: TWMSize);
var
   MinHeight: Integer;
   BtnWidth: Integer;
begin
	inherited;
   if (csDesigning in ComponentState) then
     FListBox.SetBounds(0, Height + 1, 10, 10);
   MinHeight := GetMinHeight;
   if FAllowNone then
      BtnWidth := FButton.Width * 3
   else
      BtnWidth := FButton.Width * 2;
   if Height < MinHeight then Height := MinHeight
   else begin
      if NewStyleControls then
         FBtnControl.SetBounds(ClientWidth - BtnWidth, 0, BtnWidth, ClientHeight)
      else
         FBtnControl.SetBounds(ClientWidth - BtnWidth, 1, BtnWidth, ClientHeight - 1);
      FButton.Height := FBtnControl.Height;
      FAllBtn.Height := FBtnControl.Height;
      FNoneBtn.Height := FBtnControl.Height;
      SetEditRect;
   end;
end;

procedure TAEGCheckedCombo.CMFontChanged(var Message: TMessage);
begin
   inherited;
   GetMinHeight;
end;

procedure TAEGCheckedCombo.CMEnabledChanged(var Message: TMessage);
begin
   inherited;
   FButton.Enabled := Enabled;
   FAllBtn.Enabled := Enabled;
   FNoneBtn.Enabled := Enabled;
   if Enabled then
   begin
      Color := clWindow;
      Font.Color := clWindowText;
   end
   else
   begin
      Color := clBtnFace;
      Font.Color := clGrayText;
   end;
end;

procedure TAEGCheckedCombo.CMExit(var Message: TCMExit);
begin
   inherited;
   Invalidate;
end;

procedure TAEGCheckedCombo.CMEnter(var Message: TCMGotFocus);
begin
   inherited;
   Invalidate;
end;

procedure TAEGCheckedCombo.WMSetFocus(var Message: TWMSetFocus);
begin
   inherited;
   HideCaret(Handle);
end;

procedure TAEGCheckedCombo.WMKillFocus(var Message: TWMKillFocus);
begin
   inherited;
   if FListBox.Visible then CloseUp;
end;

procedure TAEGCheckedCombo.CMCancelMode(var Message: TCMCancelMode);
begin
   with Message do
      if (Sender <> Self) and (Sender <> FBtnControl) and
         (Sender <> FButton) and (Sender <> FListBox) and
         (Sender <> FAllBtn) and (Sender <> FNoneBtn) then CloseUp;
end;

procedure TAEGCheckedCombo.WMLButtonDown(var Message: TWMLButtonDown);
begin
   NonEditMouseDown(Message);
end;

procedure TAEGCheckedCombo.WMLButtonUp(var Message: TWMLButtonUp);
begin
   MouseCapture := False;
   inherited;
end;

procedure TAEGCheckedCombo.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
   NonEditMouseDown(Message);
end;

procedure TAEGCheckedCombo.NonEditMouseDown(var Message: TWMLButtonDown);
var
   CtrlState: TControlState;
begin
   SetFocus;
   HideCaret(Handle);

   if FListBox.Visible then CloseUp
   else DropDown;

   MouseCapture := True;
   if csClickEvents in ControlStyle then
   begin
      CtrlState := ControlState;
      Include(CtrlState, csClicked);
      ControlState := CtrlState;
   end;
   with Message do
      MouseDown(mbLeft, KeysToShiftState(Keys), XPos, YPos);
end;

procedure TAEGCheckedCombo.SetDropDownLines(Value: Integer);
begin
   if FDropDownLines <> Value then
      if (Value >= MINDROPLINES) and (Value <= MAXDROPLINES) then
         FDropDownLines := Value;
end;

procedure TAEGCheckedCombo.SetColumns(Value: Integer);
begin
   if (Value >= 0) and (FColumns <> Value) then
   begin
      FColumns := Value;
      FListBox.Columns := FColumns;
   end;
end;

procedure TAEGCheckedCombo.SetSorted(Value: Boolean);
begin
   if FSorted <> Value then
   begin
      FSorted := Value;
      TStringList(FItems).Sorted := FSorted;
      FListBox.Sorted := FSorted;
   end;
end;

procedure TAEGCheckedCombo.SetAllowNone(Value: Boolean);
begin
   if FAllowNone <> Value then
   begin
      FAllowNone := Value;
      FNoneBtn.Visible := FAllowNone;
      if not FAllowNone then
      begin
         if (FCheckedCount = 0) and (FListBox.Items.Count > 0) then
            SetChecked(0, True);
      end;
      SendMessage(Handle, WM_SIZE, SIZE_RESTORED, Word(Width) or Word(Height) shl 16);
   end;
end;

procedure TAEGCheckedCombo.SetChecked(Index: Integer; Checked: Boolean);
var
   ok: Boolean;
begin
   if Index < FListBox.Items.Count then
   begin
      ok := False;
      if not FListBox.Checked[Index] and Checked then
      begin
         FCheckedCount := FCheckedCount + 1;
         ok := True;
      end
      else
         if FListBox.Checked[Index] and not Checked then
         begin
            FCheckedCount := FCheckedCount - 1;
            ok := True;
         end;
      if ok then
      begin
         FListBox.Checked[Index]:= Checked;
         UpdateText;
      end;
   end;
end;

function TAEGCheckedCombo.GetChecked(Index: Integer): Boolean;
begin
   if Index < FListBox.Items.Count then
      Result := FListBox.Checked[Index]
   else
      Result := False;
end;

procedure TAEGCheckedCombo.ClearItems;
begin
   FItems.Clear;
   FListBox.Clear;
   FCheckedCount := 0;
   UpdateText;
end;

procedure TAEGCheckedCombo.ItemsChange(Sender: TObject);
begin
   FListBox.Clear;
   FListBox.Items.Assign(FItems);
   FCheckedCount := 0;
   UpdateText;
end;

procedure TAEGCheckedCombo.ToggleOnOff(Sender: TObject);
begin
   if FListBox.Checked[FListBox.ItemIndex] then
      FCheckedCount := FCheckedCount + 1
   else
   begin
      FCheckedCount := FCheckedCount - 1;
      if (not FAllowNone) and (FCheckedCount = 0) and (FListBox.Items.Count > 0) then
      begin
         FListBox.Checked[FListBox.ItemIndex] := True;
         FCheckedCount := 1;
      end;
   end;
   UpdateText;
end;

procedure TAEGCheckedCombo.KeyDown(var Key: Word; Shift: TShiftState);
begin
   if (Key in [VK_UP, VK_DOWN, VK_LEFT, VK_RIGHT, VK_HOME, VK_END]) then
   begin
      if (Key in [VK_UP, VK_DOWN]) and (not FListBox.Visible) then
         DropDown
      else if FListBox.Items.Count > 0 then
         case Key of
            VK_UP:    if FListBox.ItemIndex > 0 then FListBox.ItemIndex := FListBox.ItemIndex - 1;
            VK_DOWN:  if FListBox.ItemIndex < FListBox.Items.Count - 1 then FListBox.ItemIndex := FListBox.ItemIndex + 1;
            VK_LEFT:  if FColumns > 0 then
                         if FListBox.ItemIndex - FActualLines >= 0 then
                            FListBox.ItemIndex := FListBox.ItemIndex - FActualLines
                         else
                            FListBox.ItemIndex := 0;
            VK_RIGHT: if FColumns > 0 then
                         if FListBox.ItemIndex + FActualLines <= FListBox.Items.Count - 1 then
                            FListBox.ItemIndex := FListBox.ItemIndex + FActualLines
                         else
                            FListBox.ItemIndex := FListBox.Items.Count - 1;
            VK_HOME:  FListBox.ItemIndex := 0;
            VK_END:   FListBox.ItemIndex := FListBox.Items.Count - 1;
         end;
      Key := 0;
   end
   else
      inherited;
end;

procedure TAEGCheckedCombo.KeyPress(var Key: Char);
begin
   if FListBox.Visible then
      if (Key = char(VK_ESCAPE)) then
         CloseUp
      else if (Key = #32) then
      begin
         FListBox.Checked[FListBox.ItemIndex] := not FListBox.Checked[FListBox.ItemIndex];
         ToggleOnOff(Self);
      end
      else
         inherited
   else
      inherited;
end;

function TAEGCheckedCombo.IsChecked(Index: Integer): Boolean;
begin
   Result := FListBox.Checked[Index];
end;

procedure TAEGCheckedCombo.LocateIndex(Index: Integer);
begin
   if FLoading then Exit;
   FLoading := True;
   if (DataField <> '') and (DataSource <> nil) and (DataSource.DataSet <> nil) and (DataSource.DataSet.Active) then
   begin
      try
         with DataSource.DataSet do
         begin
            DisableControls;
            First;
            while not EOF do
            begin
               if FieldByName(DataField).AsString = FListBox.Items[Index] then
                  Break;
               Next;
            end;
         end;
      finally
         DataSource.DataSet.EnableControls;
         FLoading := False;
      end;
   end;
end;

procedure TAEGCheckedCombo.LoadData(Checked: Boolean);
begin
   if FLoading then Exit;
   FLoading := True;
   if (DataField <> '') and (DataSource <> nil) and (DataSource.DataSet <> nil) and (DataSource.DataSet.Active) then
   begin
      try
         FDataItems.Clear;
         TStringList(FDataItems).Sorted := FSorted;
         with DataSource.DataSet do
         begin
            DisableControls;
            First;
            while not EOF do
            begin
               FDataItems.Add(FieldByName(DataField).AsString);
               Next;
            end;
            First;
            if FDataItems.Text <> FItems.Text then
            begin
               FItems.Assign(FDataItems);
               if Checked then SetCheckedAll(Self);
            end;
         end;
         FDataItems.Clear;
      finally
         DataSource.DataSet.EnableControls;
         FLoading := False;
      end;
   end;
end;

function TAEGCheckedCombo.GetText(Quote: TCHBQuoteType): String;
const
   x: array[TCHBQuoteType] of String = ('', '''', '"');
var
   s: String;
   i: Integer;
begin
   s := '';
   for i := 0 to FListBox.Items.Count-1 do
      if FListBox.Checked[i] then
         s := s + x[Quote] + FListBox.Items[i] + x[Quote] + ',';
   if s <> '' then Delete(s, Length(s), 1);  //remove trailing comma
   Result := s;
end;

function TAEGCheckedCombo.GetObjectText(Quote: TCHBQuoteType): String;
const
   x: array[TCHBQuoteType] of String = ('', '''', '"');
var
   s: String;
   i: Integer;
begin
   s := '';
   for i := 0 to FListBox.Items.Count-1 do
      if FListBox.Checked[i] then
         s := s + x[Quote] + IntToStr(LongInt(FListBox.Items.Objects[i])) + x[Quote] + ',';
   if s <> '' then Delete(s, Length(s), 1);  //remove trailing comma
   Result := s;
end;

function TAEGCheckedCombo.GetFieldText(Field: String; Quote: TCHBQuoteType): String;
const
   x: array[TCHBQuoteType] of String = ('', '''', '"');
var
   s: String;
   i: Integer;
begin
   s := '';
   for i := 0 to FListBox.Items.Count-1 do
      if FListBox.Checked[i] then
      begin
         LocateIndex(i);
         s := s + x[Quote] + DataSource.DataSet.FieldByName(Field).AsString + x[Quote] + ',';
      end;
   if s <> '' then Delete(s, Length(s), 1);  //remove trailing comma
   Result := s;
end;

procedure TAEGCheckedCombo.SetForceAll(Value: Boolean);
begin
   if FForceAll <> Value then
   begin
      FForceAll := Value;
      if FForceAll then
         CheckAll
      else
         UpdateText;
   end;
end;

procedure TAEGCheckedCombo.CheckAll;
begin
   SetCheckedAll(Self);
end;

procedure TAEGCheckedCombo.UnCheckAll;
begin
   SetUnCheckedAll(Self);
end;

procedure TAEGCheckedCombo.SetCheckedAll(Sender: TObject);
var
   i: Integer;
begin
   for i := 0 to FListBox.Items.Count-1 do
      FListBox.Checked[i] := True;
   FCheckedCount := FListBox.Items.Count;
   UpdateText;
end;

procedure TAEGCheckedCombo.SetUnCheckedAll(Sender: TObject);
var
   i: Integer;
begin
   FCheckedCount := 0;
   for i := 0 to FListBox.Items.Count-1 do
      FListBox.Checked[i] := False;
   UpdateText;
end;

procedure TAEGCheckedCombo.DropDown;
var
   P: TPoint;
   R, S: TRect;
   i, j: Double;
   Y: Integer;
   NumBtns: Integer;
   GridWidth, GridHeight: Integer;
   SysBorderWidth, SysBorderHeight: Integer;
begin
   if FListBox.Visible then Exit;

   FDropDownLines := FListBox.Items.Count;
   Columns := 0;
   if FDropDownLines > 20 then
      Columns := (FListBox.Items.Count div 20) + 1;
   if FColumns > 3 then
      Columns := 3;
   if FColumns = 1 then
      Columns := 0;
   i := FListBox.Items.Count;
   j := FColumns;
   if FColumns > 1 then
      FDropDownLines := Trunc((i / j) + 0.5);
   if FDropDownLines < MINDROPLINES then
      FDropDownLines := MINDROPLINES;
   if FDropDownLines > 20 then
      FDropDownLines := 20;

   FListBox.Font := Font;
   // Assign Form coordinate and show
   SysBorderWidth := GetSystemMetrics(SM_CXBORDER);
   SysBorderHeight := GetSystemMetrics(SM_CYBORDER);
   P := ClientOrigin;
   if NewStyleControls then
   begin
     Dec(P.X, 2 * SysBorderWidth);
     Dec(P.Y, SysBorderHeight);
   end;
   if NewStyleControls then
      Y := P.Y + ClientHeight + 3 * SysBorderHeight
   else
      Y := P.Y + Height - 1;
   if FAllowNone then
      NumBtns := 2
   else
      NumBtns := 1;
   if FColumns > 1 then
   begin
      GridWidth := (Width - (FButton.Width*NumBtns) - 2) * FColumns;
      GridHeight := FDropDownLines * FListBox.ItemHeight + 18;
   end
   else
   begin
      GridWidth := Width - (FButton.Width*NumBtns) - 2;
      GridHeight := FDropDownLines * FListBox.ItemHeight + 4;
   end;
   if (Y + GridHeight) > Screen.Height then
   begin                 
      Y := P.Y - GridHeight - 1;
      if Y < 0 then
      begin
         if NewStyleControls then
            Y := P.Y + ClientHeight + 3 * SysBorderHeight
         else
            Y := P.Y + Height - 1;
      end;
   end;
   if NewStyleControls then
      SetWindowPos(FListBox.Handle, 0, P.X, Y, GridWidth, GridHeight, SWP_NOACTIVATE)
   else
      SetWindowPos(FListBox.Handle, 0, P.X + Width - GridWidth, Y, GridWidth, GridHeight, SWP_NOACTIVATE);

   with FListBox do
   begin
      BorderStyle := bsSingle;
      Visible := True;
      FActualLines := FDropDownLines;
      if FColumns > 1 then
      begin
         R := ItemRect(0);
         for Y := FDropDownLines - 1 to Items.Count - 1 do
         begin
            S := ItemRect(Y);
            if S.Top <= R.Top then
            begin
               FActualLines := Y;
               Break;
            end;
         end;
      end;
   end;
   Invalidate;
end;

procedure TAEGCheckedCombo.CloseUp;
begin
	FListBox.Visible := False;
   SetFocus;
   Invalidate;
end;

procedure TAEGCheckedCombo.UpdateText;
var
   s: String;
   i: Integer;
begin
   if (not FAllowNone) and (FCheckedCount = 0) and (FListBox.Items.Count > 0) then
   begin
      FListBox.Checked[0] := True;
      FCheckedCount := 1;
   end;

   if FCheckedCount = 0 then
   begin
      s := 'None';
   	FSelectState := ssNone;
   end
   else if FCheckedCount = 1 then
   begin
   	for i := 0 to FListBox.Items.Count-1 do
   	begin
   		if FListBox.Checked[i] then
   		begin
            s := FListBox.Items[i];
            LocateIndex(i);
            break;
   		end;
   	end;
      FSelectState := ssOne;
   end
   else if FCheckedCount = FListBox.Items.Count then
   begin
      s := 'All';
   	FSelectState := ssAll;
   end
   else
   begin
      s := 'Multiple';
   	FSelectState := ssMultiple;
   end;

   if FForceAll then
   begin
      s := 'All';
   	FSelectState := ssAll;
   end;

	if s <> Text then
      Text := s      //causes OnChange to happen automatically
   else
	   DoChange;

   FListBox.FLastIndex := FListBox.ItemIndex;
end;

procedure TAEGCheckedCombo.DoChange;
begin
   if Assigned(OnChange) then OnChange(Self);
end;

{ TAEGComboButton }

procedure TAEGComboButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   with TAEGCheckedCombo(Parent.Parent) do
      if not FListBox.Visible then
         if (Handle <> GetFocus) and CanFocus then
         begin
            SetFocus;
            if GetFocus <> Handle then Exit;
         end;
   inherited MouseDown (Button, Shift, X, Y);
   with TAEGCheckedCombo(Parent.Parent) do
      if FListBox.Visible then CloseUp
      else DropDown;
end;

{ TAEGPopupGrid }

constructor TAEGPopupGrid.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FLastIndex := -1;
end;

procedure TAEGPopupGrid.CreateParams(var Params: TCreateParams);
begin
   inherited CreateParams(Params);
   Params.WindowClass.Style := CS_SAVEBITS;
end;

procedure TAEGPopupGrid.CreateWnd;
begin
   inherited CreateWnd;
   if not (csDesigning in ComponentState) then
      Windows.SetParent(Handle, 0);
   CallWindowProc(DefWndProc, Handle, WM_SETFOCUS, 0, 0);
end;

procedure TAEGPopupGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
   i, Index: Integer;
   InCheck: Boolean;
begin
   if Button = mbLeft then
   begin
      Index := ItemAtPos(Point(X,Y), True);
      if Index <> -1 then
      begin
         if not UseRightToLeftAlignment then
         begin
            InCheck := (X - ItemRect(Index).Left < GetCheckWidth);
         end
         else
         begin
            Dec(X, ItemRect(Index).Right - GetCheckWidth);
            InCheck := (X > 0) and (X < GetCheckWidth);
         end;
         if InCheck then
         begin
            if ssShift in Shift then
            begin
               if FLastIndex = -1 then FLastIndex := 0;
               if Self.Checked[Index] then
               begin
                  Self.Checked[Index] := False;
                  TAEGCheckedCombo(Parent).FCheckedCount := TAEGCheckedCombo(Parent).FCheckedCount - 1;
               end;
               if Index < FLastIndex then
               begin
                  for i := FLastIndex downto Index+1 do
                     if not Self.Checked[i] then
                     begin
                        Self.Checked[i] := True;
                        TAEGCheckedCombo(Parent).FCheckedCount := TAEGCheckedCombo(Parent).FCheckedCount + 1;
                     end;
               end
               else
                  if Index > FLastIndex then
                  begin
                     for i := FLastIndex to Index-1 do
                        if not Self.Checked[i] then
                        begin
                           Self.Checked[i] := True;
                           TAEGCheckedCombo(Parent).FCheckedCount := TAEGCheckedCombo(Parent).FCheckedCount + 1;
                        end;
                  end;
            end
            else if not (ssCtrl in Shift) then
            begin
               for i := 0 to Self.Items.Count-1 do
                  Self.Checked[i] := False;
               TAEGCheckedCombo(Parent).FCheckedCount := 0;
            end
         end;
      end;
      inherited MouseDown (Button, Shift, X, Y)
   end;
end;

procedure Register;
begin
   RegisterComponents('Additional', [TAEGCheckedCombo]);
end;

end.

